using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace LD10Tanks
{
    public class Bullet
    {
        public int owner_tank;
        public Vector3 pos;
        public Vector3 dir;

        public Bullet(Vector3 _pos, Vector3 _dir, int owner)
        {
            pos = _pos;
            dir = _dir;
            owner_tank = owner;
        }
    }

    public class ActiveLevel
    {
        private const float GroundDepth = 1.0f;
        private const float GameXZSize = 10.0f;
        private const float CellSize = 0.19f;

        private const float HeightNormal = 0.4f;
        private const float HeightTall = 0.65f;

        private const float BulletSpeed = 3.0f;
        private const float BulletExplosionRadius = 0.3f;

        private EvalLevel m_level;
        private BackgroundCubes m_tri_gen;
        private bool m_tri_gen_dirty;

        private ParticleManager m_particle_mgr;

        public FastList<VertexPositionNormalColor> Vertices
        {
            get { return m_tri_gen.vertices; }
        }

        public FastList<int> Triangles
        {
            get { return m_tri_gen.triangles; }
        }

        public String name = "";
        public int unique_id;

        public List<Tank> tanks = new List<Tank>();
        public List<Bullet> bullets = new List<Bullet>();

        private bool m_simulating;
        public bool SimulationStarted
        {
            get { return m_simulating; }
        }

        private float m_simulation_time;
        public float SimulationTime
        {
            get { return m_simulation_time; }
        }

        public bool SimulationFinished
        {
            get { return SimulationStarted && (bullets.Count == 0); }
        }

        public bool LevelDone
        {
            get
            {
                if (!SimulationFinished)
                    return false;
                for (int i = 0; i < tanks.Count; ++i)
                {
                    if (tanks[i].fired < tanks[i].bullets)
                        return false;
                }
                return true;
            }
        }

        public ActiveLevel(ParticleManager particle_mgr)
        {
            m_particle_mgr = particle_mgr;

            // Create the level generator
            float ground_y = -GroundDepth;
            float ground_y_top = 1.0f;
            float xz_size = GameXZSize;

            m_level = new EvalLevel();
            MarchingCubes gen = new MarchingCubes(new BoundingBox(new Vector3(-xz_size * 0.5f - CellSize * 3, ground_y - CellSize * 3, -xz_size * 0.5f - CellSize * 3), new Vector3(xz_size * 0.5f + CellSize * 3, ground_y_top + CellSize * 3, xz_size * 0.5f + CellSize * 3)), CellSize);
            m_tri_gen = new BackgroundCubes(gen);

            // Initialize a sphere grid to the exlusion fields (explosion generate a lot of spheres)
            Vector2 grid_extent = new Vector2(6, 6);
            float grid_cell = 0.5f;

            //m_level.FieldExclude.InitSphereGrid(-grid_extent, grid_extent, grid_cell);
        }

        public void Quit()
        {
            m_tri_gen.Close();
        }

        public void Reset(GameLevel level)
        {
            // Misc inits
            name = level.name;
            unique_id = level.unique_id;

            // Add the tanks
            tanks.Clear();
            foreach (Tank tank in level.tanks)
            {
                tanks.Add(new Tank(tank));
            }

            // Create the level
            Revert(level);
        }

        public void Revert(GameLevel level)
        {
            m_level.Clear();

            // Add the ground
            float ground_y = -GroundDepth;
            float xz_size = GameXZSize;
            m_level.FieldSolid.AddBox(new BoundingBox(new Vector3(-xz_size * 0.5f, ground_y, -xz_size * 0.5f), new Vector3(xz_size * 0.5f, 0.0f, xz_size * 0.5f)));
            m_level.StartPoints.Add(new Vector3(0, ground_y + CellSize, 0));

            // Add the walls
            foreach (Wall wall in level.walls)
            {
                EvalField dest = null;
                float y = 0;
                float height = 0;

                switch (wall.type)
                {
                    case WallType.Normal:
                        dest = m_level.FieldSolid;
                        height = HeightNormal;
                        y = height * 0.5f;
                        break;

                    case WallType.Tall:
                        dest = m_level.FieldSolid;
                        height = HeightTall;
                        y = height * 0.5f;
                        break;

                    case WallType.HoleShallow:
                        {
                            dest = m_level.FieldExclude;
                            height = HeightTall + HeightNormal + 0.1f;
                            y = -HeightNormal + height * 0.5f;
                        }
                        break;

                    case WallType.HoleDeep:
                        {
                            dest = m_level.FieldExclude;
                            height = HeightTall + HeightNormal + 0.1f;
                            y = -HeightTall + height * 0.5f;
                        }
                        break;
                }

                if (dest != null)
                    AddWall(dest, y, height, wall.pos1, wall.pos2, wall.thickness);
            }

            // Start generating the field
            m_tri_gen.StartJob(m_level);

            // Clear the simulation flags
            bullets.Clear();
            m_simulating = false;
            m_simulation_time = 0;

            // Reset the tanks
            for (int i = 0; i < tanks.Count; ++i)
                tanks[i].Reset();

            // Wait for the field to finish
            m_tri_gen.SyncJob();
            m_tri_gen_dirty = false;
        }

        public void StartSimulation(int start_tank)
        {
            if (m_simulating)
                return;

            m_simulating = true;

            // Fire the first tank
            FireTank(start_tank);
        }

        public void Update(float delta_t)
        {
            // Update the geometry generation
            if (m_tri_gen.JobStarted && !m_tri_gen.JobRunning)
            {
                m_tri_gen.SyncJob();
            }

            if (m_tri_gen_dirty && !m_tri_gen.JobStarted)
            {
                m_tri_gen.StartJob(m_level);
                m_tri_gen_dirty = false;
            }

            // Update the world
            if (!SimulationStarted || SimulationFinished)
                return;

            // Count time
            m_simulation_time += delta_t;

            // Update the bullets
            {
                int idx = 0;
                while (idx < bullets.Count)
                {
                    Bullet bullet = bullets[idx];
                    bool kill = false;

                    // Move
                    bullet.pos += bullet.dir * (delta_t * BulletSpeed);

                    // Hit a wall?
                    if (m_level.IsInside(ref bullet.pos))
                    {
                        // Eat a bit away
                        m_level.FieldExclude.AddSphere(bullet.pos, BulletExplosionRadius);

                        // Add particle effects
                        m_particle_mgr.SpawnParticles(
                            bullet.pos,                 // Position
                            0.3f,                       // Radius
                            0.3f,                       // Speed
                            1.5f,                       // Lifetime
                            500,                        // Amount
                            new Color(58, 71, 24, 250));

                        // Kill this bullet
                        kill = true;
                        // Generate the level later
                        m_tri_gen_dirty = true;
                    }

                    // Hit a tank?
                    if (!kill)
                    {
                        int tank = GetTankIdx(bullet.pos, bullet.owner_tank, 0.3f);
                        if (tank != -1)
                        {
                            FireTank(tank);
                            kill = true;
                        }
                    }

                    // Out of bounds?
                    if (!kill)
                    {
                        if (bullet.pos.X < -5.5f || bullet.pos.Z < -5.5f ||
                            bullet.pos.X > 5.5f || bullet.pos.Z > 5.5f)
                        {
                            kill = true;
                        }
                    }

                    // Kill the bullet?
                    if (kill)
                        bullets.RemoveAt(idx);
                    else
                        ++idx;
                }
            }
        }

        /// <summary>
        /// Returns the tank index the point hits or -1 if none.
        /// </summary>
        public int GetTankIdx(Vector3 world_pos, int ignore_tank, float dist)
        {
            float req_dist2 = dist * dist;

            Vector2 pos = new Vector2(world_pos.X, world_pos.Z);
            for (int i = 0; i < tanks.Count; ++i)
            {
                if (i == ignore_tank)
                    continue;

                float dist2 = (pos - tanks[i].pos).LengthSquared();
                if (dist2 <= req_dist2)
                    return i;
            }

            return -1;
        }

        /// <summary>
        /// Tests if the tank intersects other tanks.
        /// </summary>
        public bool TankIntersectsTanks(int idx)
        {
            float tank_rad = 0.3f;
            float req_rad = tank_rad * 2.0f;
            float req_rad2 = req_rad * req_rad;

            Tank tank = tanks[idx];
            for (int i = 0; i < tanks.Count; ++i)
            {
                if (i == idx)
                    continue;

                Vector2 dir = tanks[i].pos - tank.pos;
                if (dir.LengthSquared() <= req_rad2)
                    return true;
            }
            return false;
        }

        /// <summary>
        /// Fires the given tank.
        /// </summary>
        private void FireTank(int tank_idx)
        {
            if (tank_idx < 0)
                return;
            Tank tank = tanks[tank_idx];
            if (tank.fired >= tank.bullets)
                return;

            Vector3 shoot_pos = tank.GunPos3D;
            for (int i = 0; i < tank.GunCount; ++i)
            {
                Vector3 dir = tank.GetGunDir(i);
                bullets.Add(new Bullet(shoot_pos, dir, tank_idx));

                // Add particles (smoke)
                m_particle_mgr.SpawnParticles(
                    shoot_pos + dir * 0.4f,     // Position
                    0.1f,                       // Radius
                    0.1f,                       // Speed
                    0.5f,                       // Lifetime
                    50,                         // Amount
                    new Color(48, 61, 14, 150));
            }
            tank.fired++;
        }

        private void AddWall(EvalField dest, float y, float height, Vector2 pt1, Vector2 pt2, float thickness)
        {
            Vector3 dir = new Vector3(pt2.X, 0, pt2.Y) - new Vector3(pt1.X, 0, pt1.Y);
            Vector3 up = new Vector3(0, 1, 0);

            Vector2 center_2d = Vector2.Lerp(pt1, pt2, 0.5f);
            Vector3 center = new Vector3(center_2d.X, y, center_2d.Y);

            float dir_len = dir.Length();
            float up_len = Math.Abs(height);
            float side_len = thickness;

            dest.AddBox(center, dir, up, dir_len, up_len, side_len);
        }
    }
}
