uniform extern float4x4 World;
uniform extern float4x4 View;
uniform extern float4x4 Proj;
uniform extern float4x4 WorldViewProj;

uniform extern float4 LightColor;
uniform extern float3 LightDir;
uniform extern float3 LightPos;

uniform extern texture Texture1;
uniform extern texture Texture2;

sampler texture1_sampler = sampler_state
{
    AddressU = Clamp;
    AddressV = Clamp;
	Texture = <Texture1>;
	minfilter = POINT;
	magfilter = LINEAR;
	mipfilter = LINEAR;
};
sampler texture2_sampler = sampler_state
{
    AddressU = Clamp;
    AddressV = Clamp;
	Texture = <Texture2>;
	minfilter = POINT;
	magfilter = LINEAR;
	mipfilter = LINEAR;
};

struct VS_OUTPUT
{
    float4 position : POSITION;
    float4 color : COLOR0;
};

struct DEPTH_VS_OUTPUT
{
    float4 position : POSITION;
    float2 depth : TEXCOORD0;
};

struct AO_VS_OUTPUT
{
    float4 position : POSITION;
    float2 uv : TEXCOORD0;
};

static const int blur_kernel_size = 15;
float2 PixelKernelH[blur_kernel_size];
float2 PixelKernelW[blur_kernel_size];
float BlurWeights[blur_kernel_size];

static const int ao_neighbor_size = 12;
float2 AOSamples[ao_neighbor_size];

VS_OUTPUT NormalTransform(
    float4 pos : POSITION,
    float3 normal : NORMAL,
    float4 color : COLOR0 )
{
    VS_OUTPUT ret = (VS_OUTPUT)0;

	// Rotate normal according to the world matrix
    normal = mul(normal, (float3x3)World);

    ret.position = mul(pos, WorldViewProj);
    
//    float light = -dot(LightDir, normal);

	float3 ldir = normalize(mul(pos, World) - LightPos);
	float light = -dot(ldir, normal);
    
//    light = abs(light); // Two sided lighting
    
    ret.color = color * LightColor * float4(light, light, light, 1);
    ret.color += float4(0.1, 0.1, 0.1, 0); // Ambient term
//	ret.color = float4(normal, 1);

    return ret;
}

VS_OUTPUT AlphaTransform(
    float4 pos : POSITION,
    float4 color : COLOR0 )
{
	VS_OUTPUT ret = (VS_OUTPUT)0;
	
	ret.position = mul(pos, WorldViewProj);
	ret.color = color;
	return ret;
}

DEPTH_VS_OUTPUT DepthTransform(float4 pos : POSITION)
{
	DEPTH_VS_OUTPUT ret = (DEPTH_VS_OUTPUT)0;
	
	float4 view_pos = mul(mul(pos, World), View);
	ret.position = mul(view_pos, Proj);
	ret.depth = -view_pos.z;
//	ret.depth = ret.position.z;
//	ret.depth = ret.position.z / ret.position.w;
	return ret;
}

AO_VS_OUTPUT ApplyAOTransform(
	float4 pos : POSITION,
	float2 uv : TEXCOORD0)
{
	AO_VS_OUTPUT ret = (AO_VS_OUTPUT)0;
	
	ret.position = pos;
	ret.uv = uv;
	return ret;
}

float4 NormalPixelShader(VS_OUTPUT vsout) : COLOR
{
	return vsout.color;
}

float4 AlphaPixelShader(VS_OUTPUT vsout) : COLOR
{
	return vsout.color;
}

float4 DepthPixelShader(DEPTH_VS_OUTPUT vsout) : COLOR
{
	float c = vsout.depth.x;
	return float4(c, c, c, 1);
}

float4 ApplyAOPixelShader(AO_VS_OUTPUT vsout) : COLOR
{
//	float intensity = tex2D(texture1_sampler, vsout.uv).r * 0.1f;
//	return float4(intensity, intensity, intensity, 1);

	float orig_depth = tex2D(texture1_sampler, vsout.uv).r;
	
	// Check the neighbors for maximum z distance
/*	float z_dist = 0;
	for (int i = 0; i < ao_neighbor_size; ++i)
	{
		float ndist = abs(tex2D(texture1_sampler, vsout.uv + AOSamples[i]).r - orig_depth);
//		z_dist = max(z_dist, ndist);
		z_dist += ndist;
	}
	z_dist /= ao_neighbor_size;*/
	
	// Get the applying factor from the z_dist (the bigger the "jump", the less we want to blur)
	static const float apply_mul = 1.0f;
//	float apply_mul = 1.0f * clamp(1.0f - z_dist * 0.1f, 0, 1);
	
	// Get the blurred depth and the distance between it and the original
	float blur_depth = tex2D(texture2_sampler, vsout.uv).r;
//	float diff = abs(orig_depth - blur_depth) / orig_depth;
//	float diff = abs(orig_depth - blur_depth);
	float diff = max(orig_depth - blur_depth, 0);
	
	float ret = clamp(1 - diff * apply_mul, 0, 1);
//	return ret;
	return ret * ret * ret;
//return orig_depth * 10.0f;
//return blur_depth * 10.0f;
//return abs(diff) * 100.0f;
//return z_dist;
//return apply_mul;
}

float4 GaussianBlurWPixelShader(AO_VS_OUTPUT vsout) : COLOR
{
	float4 col = 0;
	
	for (int i = 0; i < blur_kernel_size; ++i)
	{
		col += tex2D(texture1_sampler, vsout.uv + PixelKernelW[i].xy) * BlurWeights[i];
	}
	return col;
}

float4 GaussianBlurHPixelShader(AO_VS_OUTPUT vsout) : COLOR
{
	float4 col = 0;
	
	for (int i = 0; i < blur_kernel_size; ++i)
	{
		col += tex2D(texture1_sampler, vsout.uv + PixelKernelH[i].xy) * BlurWeights[i];
	}
	return col;
}

technique NormalRenderTechnique
{
    pass P0
    {
        vertexShader = compile vs_2_0 NormalTransform();
        pixelShader = compile ps_2_0 NormalPixelShader();
        ZEnable = true;
        AlphaBlendEnable = false;
    }
}

technique AlphaRenderTechnique
{
    pass P0
    {
        vertexShader = compile vs_2_0 AlphaTransform();
        pixelShader = compile ps_2_0 AlphaPixelShader();
        ZEnable = true;
        AlphaBlendEnable = true;
    }
}

technique DepthRenderTechnique
{
	pass P0
	{
        vertexShader = compile vs_2_0 DepthTransform();
        pixelShader = compile ps_2_0 DepthPixelShader();
        ZEnable = true;
        AlphaBlendEnable = false;
	}
}

technique ApplyAOTechnique
{
	pass P0
	{
        vertexShader = compile vs_2_0 ApplyAOTransform();
        pixelShader = compile ps_2_0 ApplyAOPixelShader();
        ZEnable = false;
        AlphaBlendEnable = true;
	}
}

technique GaussianBlurWTechnique
{
	pass P0
	{
        vertexShader = compile vs_2_0 ApplyAOTransform();
        pixelShader = compile ps_2_0 GaussianBlurWPixelShader();
        ZEnable = false;
        AlphaBlendEnable = false;
	}
}

technique GaussianBlurHTechnique
{
	pass P0
	{
        vertexShader = compile vs_2_0 ApplyAOTransform();
        pixelShader = compile ps_2_0 GaussianBlurHPixelShader();
        ZEnable = false;
        AlphaBlendEnable = false;
	}
}
