// Define this to open all levels
//#define CHEAT

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Storage;

namespace LD10Tanks
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class GameMain : Microsoft.Xna.Framework.Game
    {
        private enum ControlMode
        {
            None,
            MoveCamera,
            MoveTank,
            AimTank,
            PickStartTank,
        }

        private const string Version = "v1.10";

        private GraphicsDeviceManager graphics;
        private ContentManager content;

        private Highscores m_highscores;
        private GameLevelList m_levels;
        private ActiveLevel m_active_level;
        private int m_current_level;

        private BasicEffect m_basic_effect;
        private Effect m_normal_shader;
        private SpriteFont m_font;
        private SpriteFont m_font_medium;
        private SpriteBatch m_font_batch;
        private VertexDeclaration m_vdecl_VertexPositionColor;
        private VertexDeclaration m_vdecl_VertexPositionNormalColor;
        private VertexDeclaration m_vdecl_VertexPositionTexture;

        private bool m_ao_enabled = true;

        private ParticleManager m_particle_mgr;
        private TankMeshes m_tank_meshes;
        private BackgroundMesh m_bg_mesh;

        public static Random random = new Random();

        public GameMain()
        {
            graphics = new GraphicsDeviceManager(this);
            content = new ContentManager(Services);
        }

        protected override void OnExiting(object sender, EventArgs args)
        {
            m_active_level.Quit();

            base.OnExiting(sender, args);
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            base.Initialize();

            m_tank_meshes = new TankMeshes();
            m_bg_mesh = new BackgroundMesh();

            m_particle_mgr = new ParticleManager();

            m_highscores = new Highscores();
            m_levels = new GameLevelList();
            m_active_level = new ActiveLevel(m_particle_mgr);

            // Pick the last unplayed level
            m_current_level = 0;
            while (m_current_level < (m_levels.levels.Count - 1))
            {
                if (m_highscores.GetLevelScore(m_levels.levels[m_current_level].unique_id) < 0)
                    break;
                ++m_current_level;
            }
            m_active_level.Reset(m_levels.levels[m_current_level]);
        }

        private RenderTarget2D m_render_target_depth;
        private RenderTarget2D m_render_target_depth_blur_tmp;
        private RenderTarget2D m_render_target_depth_blur;

        /// <summary>
        /// Load your graphics content.  If loadAllContent is true, you should
        /// load content from both ResourceManagementMode pools.  Otherwise, just
        /// load ResourceManagementMode.Manual content.
        /// </summary>
        /// <param name="loadAllContent">Which type of content to load.</param>
        protected override void LoadGraphicsContent(bool loadAllContent)
        {
            if (loadAllContent)
            {
                m_basic_effect = new BasicEffect(graphics.GraphicsDevice, null);
                m_normal_shader = content.Load<Effect>("Content/NormalShader");
                m_font = content.Load<SpriteFont>("Content/MainFont");
                m_font_medium = content.Load<SpriteFont>("Content/MediumFont");
                m_font_batch = new SpriteBatch(graphics.GraphicsDevice);
                m_vdecl_VertexPositionColor = new VertexDeclaration(graphics.GraphicsDevice, VertexPositionColor.VertexElements);
                m_vdecl_VertexPositionNormalColor = new VertexDeclaration(graphics.GraphicsDevice, VertexPositionNormalColor.VertexElements);
                m_vdecl_VertexPositionTexture = new VertexDeclaration(graphics.GraphicsDevice, VertexPositionTexture.VertexElements);

                int scr_w = graphics.GraphicsDevice.PresentationParameters.BackBufferWidth;
                int scr_h = graphics.GraphicsDevice.PresentationParameters.BackBufferHeight;

                m_ao_enabled =
                    GraphicsAdapter.DefaultAdapter.CheckDeviceFormat(
                        DeviceType.Hardware,
                        graphics.GraphicsDevice.DisplayMode.Format,
                        ResourceUsage.ResolveTarget,
                        QueryUsages.PostPixelShaderBlending,
                        ResourceType.RenderTarget,
                        SurfaceFormat.Single);

                if (m_ao_enabled)
                {
                    m_render_target_depth = new RenderTarget2D(
                        graphics.GraphicsDevice,
                        scr_w,
                        scr_h,
                        1,
                        SurfaceFormat.Single);

                    m_render_target_depth_blur_tmp = new RenderTarget2D(
                        graphics.GraphicsDevice,
                        scr_w,
                        scr_h,
                        1,
                        SurfaceFormat.Single);

                    m_render_target_depth_blur = new RenderTarget2D(
                        graphics.GraphicsDevice,
                        scr_w,
                        scr_h,
                        1,
                        SurfaceFormat.Single);
                }

                // Make the blurring coefficients
                int blur_kernel_size = m_normal_shader.Parameters["BlurWeights"].Elements.Count;

                Vector2[] pixel_kernel = new Vector2[blur_kernel_size];
                for (int i = 0; i < blur_kernel_size; ++i)
                    pixel_kernel[i] = new Vector2((float)(i - blur_kernel_size / 2) / (float)scr_w, 0);
                m_normal_shader.Parameters["PixelKernelW"].SetValue(pixel_kernel);

                for (int i = 0; i < blur_kernel_size; ++i)
                    pixel_kernel[i] = new Vector2(0, (float)(i - blur_kernel_size / 2) / (float)scr_h);
                m_normal_shader.Parameters["PixelKernelH"].SetValue(pixel_kernel);

                float[] blur_weight = new float[blur_kernel_size];
                float std_dev = 3.0f; // Blur amount
                float gaussian_const = 2.0f * std_dev * std_dev;
                float gaussian_mul = 1.0f / (float)Math.Sqrt(gaussian_const * (float)Math.PI);
                for (int i = 0; i < blur_kernel_size; ++i)
                {
                    float r = (float)(i - blur_kernel_size / 2);
                    blur_weight[i] = gaussian_mul * (float)Math.Pow(Math.E, -(r * r) / gaussian_const);
                }
                m_normal_shader.Parameters["BlurWeights"].SetValue(blur_weight);

                // Make the AO neighbor texel coords
                int ao_sample_size = m_normal_shader.Parameters["AOSamples"].Elements.Count;
                Vector2[] ao_samples = new Vector2[ao_sample_size];

                float ao_neighbor_size = 2.0f;

                for (int i = 0; i < ao_sample_size; ++i)
                {
                    float x = (float)random.NextDouble() * ao_neighbor_size;
                    float y = (float)random.NextDouble() * ao_neighbor_size;
                    if (random.Next(2) == 0) x = -x;
                    if (random.Next(2) == 0) y = -y;

                    ao_samples[i] = new Vector2(x / (float)scr_w, y / (float)scr_h);
                }
/*                {
                    int idx = 0;
                    for (int y = -1; y <= 1; ++y)
                    {
                        for (int x = -1; x <= 1; ++x)
                        {
                            if (x == 0 && y == 0)
                                continue;
                            ao_samples[idx++] = new Vector2((float)x / (float)scr_w, (float)y / (float)scr_h);
                        }
                    }
                }*/

                m_normal_shader.Parameters["AOSamples"].SetValue(ao_samples);
            }
        }


        /// <summary>
        /// Unload your graphics content.  If unloadAllContent is true, you should
        /// unload content from both ResourceManagementMode pools.  Otherwise, just
        /// unload ResourceManagementMode.Manual content.  Manual content will get
        /// Disposed by the GraphicsDevice during a Reset.
        /// </summary>
        /// <param name="unloadAllContent">Which type of content to unload.</param>
        protected override void UnloadGraphicsContent(bool unloadAllContent)
        {
            if (unloadAllContent)
            {
                content.Unload();
            }
        }

        private float m_cam_ang_y = 0.0f;
        private float m_cam_ang_x = MathHelper.ToRadians(40);

        private int m_last_mouse_x, m_last_mouse_y;
        private Vector3 m_last_mouse_field;
        private bool m_last_mouse_left;
        private bool m_last_mouse_right;

        private bool m_last_tab;

        private ControlMode m_cam_mode = ControlMode.MoveCamera;
        private int m_moving_tank_idx = -1;
        private Vector2 m_moving_tank_orig_pos;

        private bool m_sim_end_seen;
        private bool m_next_level_on_restart;

        private bool m_do_ao = true;

        private string m_center_note1;
        private string m_center_note2;
        private float m_center_note_timer;

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input and playing audio.
        /// </summary>
        /// <param name="time">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime time)
        {
            // Debug keys
            m_do_ao = !Keyboard.GetState().IsKeyDown(Keys.F12);

            // Project mouse onto the game plane
            Vector3 mouse_game_pos = Vector3.Zero;

            {
                Vector3 near_scr = new Vector3(Mouse.GetState().X, Mouse.GetState().Y, 0);
                Vector3 far_scr = new Vector3(Mouse.GetState().X, Mouse.GetState().Y, 1);
                Matrix world = Matrix.Identity;

                Vector3 ray_near = graphics.GraphicsDevice.Viewport.Unproject(near_scr, m_proj_mtx, m_view_mtx, world);
                Vector3 ray_far = graphics.GraphicsDevice.Viewport.Unproject(far_scr, m_proj_mtx, m_view_mtx, world);
                Vector3 ray_dir = ray_far - ray_near;
                ray_dir.Normalize();

                // Do a ray-plane intersection (Ray class seems to use boxed nullable types)
                float denom = Vector3.Dot(ray_dir, Vector3.UnitY);
                float plane_height = 0.0f;
                if (Math.Abs(denom) > 0.001f)
                {
                    float t = -(Vector3.Dot(ray_near, Vector3.UnitY) - plane_height) / denom;
                    if (t >= 0)
                    {
                        mouse_game_pos = ray_near + ray_dir * t;
                    }
                }
            }

            // Get the mouse/keyboard keys
            bool mouse_left = (Mouse.GetState().LeftButton == ButtonState.Pressed);
            bool mouse_right = (Mouse.GetState().RightButton == ButtonState.Pressed);

            bool tab = Keyboard.GetState().IsKeyDown(Keys.Tab);
            bool shift = Keyboard.GetState().IsKeyDown(Keys.LeftShift) || Keyboard.GetState().IsKeyDown(Keys.RightShift);

            if (!IsActive)
            {
                mouse_left = false;
                mouse_right = false;
                tab = false;
                shift = false;
            }

            // The destination control/camera mode
            ControlMode old_cam_mode = m_cam_mode;
            ControlMode new_cam_mode = ControlMode.MoveCamera;

            if (Keyboard.GetState().IsKeyDown(Keys.D1))
                new_cam_mode = ControlMode.MoveTank;
            else if (Keyboard.GetState().IsKeyDown(Keys.D2))
                new_cam_mode = ControlMode.AimTank;
            else if (Keyboard.GetState().IsKeyDown(Keys.D3))
                new_cam_mode = ControlMode.PickStartTank;

            if (m_active_level.SimulationStarted)
                new_cam_mode = ControlMode.MoveCamera;

            if (m_cam_mode != new_cam_mode)
            {
                // Switch to the new mode
                m_cam_mode = new_cam_mode;
                mouse_left = false;
                mouse_right = false;
            }

            // Changing levels
            if (!m_active_level.SimulationStarted)
            {
                if (!m_last_tab && tab && !shift && CanMoveLevelIdx(1))
                    MoveLevelIdx(1);
                if (!m_last_tab && tab && shift && CanMoveLevelIdx(-1))
                    MoveLevelIdx(-1);
            }

            // Mouse control
            if (mouse_left)
            {
                const float tank_pick_radius = 0.3f;

                // Mode specific tests
                switch (m_cam_mode)
                {
                    case ControlMode.MoveCamera:
                        IsMouseVisible = false;

                        // Camera tracking
                        int dx = Mouse.GetState().X - m_last_mouse_x;
                        int dy = Mouse.GetState().Y - m_last_mouse_y;
                        Mouse.SetPosition(m_last_mouse_x, m_last_mouse_y);

                        // Camera moving
                        float move_spd = 0.007f;
                        m_cam_ang_x = MathHelper.Clamp(m_cam_ang_x + (float)dy * move_spd, MathHelper.ToRadians(10.0f), MathHelper.ToRadians(90.0f));
                        m_cam_ang_y += (float)dx * move_spd;
                        break;

                    case ControlMode.MoveTank:
                        if (!m_last_mouse_left)
                        {
                            m_moving_tank_idx = m_active_level.GetTankIdx(mouse_game_pos, -1, tank_pick_radius);

                            // Do not move fixed tanks
                            if (m_moving_tank_idx != -1 && m_active_level.tanks[m_moving_tank_idx].fixed_pos)
                                m_moving_tank_idx = -1;

                            // Get the original position
                            if (m_moving_tank_idx != -1)
                            {
                                m_moving_tank_orig_pos = m_active_level.tanks[m_moving_tank_idx].pos;
                            }
                        }
                        if (m_moving_tank_idx != -1)
                        {
                            Tank tank = m_active_level.tanks[m_moving_tank_idx];
                            Vector3 move_amt = mouse_game_pos - m_last_mouse_field;
                            tank.pos.X = MathHelper.Clamp(tank.pos.X + move_amt.X, -5 + 0.5f, 5 - 0.5f);
                            tank.pos.Y = MathHelper.Clamp(tank.pos.Y + move_amt.Z, -5 + 0.5f, 5 - 0.5f);
                        }
                        break;

                    case ControlMode.AimTank:
                        if (!m_last_mouse_left)
                        {
                            m_moving_tank_idx = m_active_level.GetTankIdx(mouse_game_pos, -1, tank_pick_radius);
                        }
                        if (m_moving_tank_idx != -1)
                        {
                            Tank tank = m_active_level.tanks[m_moving_tank_idx];
                            Vector3 aim_dir = mouse_game_pos - tank.Pos3D;
                            aim_dir.Y = 0;

                            if (aim_dir.LengthSquared() > 0.05f)
                            {
                                aim_dir.Normalize();
                                tank.look_dir = aim_dir;
                            }
                        }
                        break;

                    case ControlMode.PickStartTank:
                        if (!m_last_mouse_left)
                        {
                            m_moving_tank_idx = m_active_level.GetTankIdx(mouse_game_pos, -1, tank_pick_radius);
                            if (m_moving_tank_idx != -1)
                            {
                                m_active_level.StartSimulation(m_moving_tank_idx);
                            }
                        }
                        break;

                    case ControlMode.None:
                        break;
                }
            }
            else
            {
                IsMouseVisible = true;

                if (old_cam_mode == ControlMode.MoveTank && m_moving_tank_idx != -1)
                {
                    if (IsTankAtValidPlace(m_moving_tank_idx) == false)
                    {
                        m_active_level.tanks[m_moving_tank_idx].pos = m_moving_tank_orig_pos;
                    }
                }
                m_moving_tank_idx = -1;
            }

            if (mouse_right)
            {
                if (m_active_level.SimulationStarted && !m_last_mouse_right)
                {
                    // Restart a new or same level?
                    if (m_next_level_on_restart)
                    {
                        MoveLevelIdx(1);
                    }
                    else
                    {
                        m_active_level.Revert(m_levels.levels[m_current_level]);
                    }

                    m_next_level_on_restart = false;
                    m_sim_end_seen = false;
                    m_center_note_timer = 0.0f;
                }
            }

            m_last_mouse_x = Mouse.GetState().X;
            m_last_mouse_y = Mouse.GetState().Y;
            m_last_mouse_field = mouse_game_pos;
            m_last_mouse_left = mouse_left;
            m_last_mouse_right = mouse_right;

            m_last_tab = tab;

            // Update the view and projection matrices
            m_view_mtx = Matrix.CreateRotationY(m_cam_ang_y) * Matrix.CreateRotationX(m_cam_ang_x) * Matrix.CreateTranslation(0, 0, -7);
            m_proj_mtx = Matrix.CreatePerspectiveFieldOfView(
                MathHelper.ToRadians(80),
                (float)graphics.GraphicsDevice.Viewport.Width / (float)graphics.GraphicsDevice.Viewport.Height,
                0.1f, 50.0f);

            // Update the active level
            m_active_level.Update((float)time.ElapsedGameTime.TotalSeconds);

            // Update the particles
            m_particle_mgr.Update((float)time.ElapsedGameTime.TotalSeconds);

            // Simulation end tracking
            if (m_active_level.SimulationStarted &&
                m_sim_end_seen == false &&
                m_active_level.SimulationFinished)
            {
                m_sim_end_seen = true;

                // Do appropriate the end level actions
                m_center_note_timer = 3.0f;
                if (m_active_level.LevelDone)
                {
                    m_center_note1 = "Level complete!";
                    if (m_highscores.RecordScore(m_active_level.unique_id, m_active_level.SimulationTime))
                    {
                        m_center_note2 = "Congratulations! A new best time!";

                        m_highscores.Save();
                        m_next_level_on_restart = true;
                    }
                }
                else
                {
                    m_center_note1 = "Try again";
                    m_center_note2 = "Not all cannons fired..";
                }
            }

            // On screen note tracking
            if (m_center_note_timer > 0.0f)
            {
                m_center_note_timer -= (float)time.ElapsedGameTime.TotalSeconds;
            }
            else
            {
                m_center_note1 = null;
                m_center_note2 = null;
            }

            base.Update(time);
        }

        private bool CanMoveLevelIdx(int level_idx_add)
        {
            int new_idx = m_current_level + level_idx_add;
            if (new_idx < 0 || new_idx >= m_levels.levels.Count)
                return false;
            if (new_idx == 0)
                return true;

            // If a cheat mode is active, then all levels are open
#if CHEAT
            return true;
#else
            // Either the current level or the new level must have a valid highscore
            return
                (m_highscores.GetLevelScore(m_levels.levels[m_current_level].unique_id) >= 0.0f) ||
                (m_highscores.GetLevelScore(m_levels.levels[new_idx].unique_id) >= 0.0f);
#endif
        }

        /// <summary>
        /// Moves to a next/prev level if it's open.
        /// </summary>
        private void MoveLevelIdx(int level_idx_add)
        {
            if (!CanMoveLevelIdx(level_idx_add))
                return;

            m_current_level += level_idx_add;
            m_active_level.Reset(m_levels.levels[m_current_level]);
        }

        private bool IsTankAtValidPlace(int tank_idx)
        {
            if (m_active_level.TankIntersectsTanks(tank_idx))
                return false;

            Vector2 pos = m_active_level.tanks[tank_idx].pos;
            return !m_levels.levels[m_current_level].BallIntersectsWalls(pos, 0.3f);
        }

        private Matrix m_view_mtx = Matrix.Identity;
        private Matrix m_proj_mtx = Matrix.Identity;

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            Vector3 light_dir = new Vector3(1.0f, -1.0f, 1.0f);
            light_dir.Normalize();

            Vector3 light_pos = new Vector3(0.0f, 1.0f, 0.0f);

            m_normal_shader.Parameters["LightDir"].SetValue(light_dir);
            m_normal_shader.Parameters["LightPos"].SetValue(light_pos);
            m_normal_shader.Parameters["View"].SetValue(m_view_mtx);
            m_normal_shader.Parameters["Proj"].SetValue(m_proj_mtx);

            // Render the depth buffer
            if (m_ao_enabled)
            {
                graphics.GraphicsDevice.SetRenderTarget(0, m_render_target_depth);
                graphics.GraphicsDevice.Clear(Color.Black);

                m_normal_shader.CurrentTechnique = m_normal_shader.Techniques["DepthRenderTechnique"];
                RenderScene(graphics);
                graphics.GraphicsDevice.ResolveRenderTarget(0);

                // Blur the depth buffer
                graphics.GraphicsDevice.SetRenderTarget(0, m_render_target_depth_blur_tmp);
                m_normal_shader.CurrentTechnique = m_normal_shader.Techniques["GaussianBlurWTechnique"];
                m_normal_shader.Parameters["Texture1"].SetValue(m_render_target_depth.GetTexture());
                RenderQuad(graphics);
                graphics.GraphicsDevice.ResolveRenderTarget(0);

                graphics.GraphicsDevice.SetRenderTarget(0, m_render_target_depth_blur);
                m_normal_shader.CurrentTechnique = m_normal_shader.Techniques["GaussianBlurHTechnique"];
                m_normal_shader.Parameters["Texture1"].SetValue(m_render_target_depth_blur_tmp.GetTexture());
                RenderQuad(graphics);
                graphics.GraphicsDevice.ResolveRenderTarget(0);
            }

            // Render the normal screen
            graphics.GraphicsDevice.SetRenderTarget(0, null);
            graphics.GraphicsDevice.Clear(Color.CornflowerBlue);
            m_normal_shader.CurrentTechnique = m_normal_shader.Techniques["NormalRenderTechnique"];
            RenderScene(graphics);

            // Render the alpha effects
            RenderSceneEffects(graphics);

            // Render the post processing
            if (m_ao_enabled)
            {
                m_normal_shader.CurrentTechnique = m_normal_shader.Techniques["ApplyAOTechnique"];
                m_normal_shader.Parameters["Texture1"].SetValue(m_render_target_depth.GetTexture());
                m_normal_shader.Parameters["Texture2"].SetValue(m_render_target_depth_blur.GetTexture());
                m_normal_shader.CommitChanges();

                graphics.GraphicsDevice.RenderState.BlendFunction = BlendFunction.Add;
                graphics.GraphicsDevice.RenderState.SourceBlend = Blend.Zero;
                graphics.GraphicsDevice.RenderState.DestinationBlend = Blend.SourceColor;
                if (m_do_ao)
                {
                    RenderQuad(graphics);
                }
            }

            // Render the UI
            {
                float h = m_font.LineSpacing;
                float h2 = m_font_medium.LineSpacing;

                // Show the title
                Vector2 pos = new Vector2(30, 20);
                m_font_batch.Begin();
                pos.Y += DrawString(m_font, "Level #" + (m_current_level + 1) + ": " + m_active_level.name, pos, Color.White);
                pos.Y += 10;

                Vector2 right_pos = pos;

                // Show the keyboard info
                if (!m_active_level.SimulationStarted)
                {
                    pos.Y += DrawString(m_font_medium, "Hold [1/2/3] to move/aim/fire a cannon", pos, Color.LightGray);

                    if (CanMoveLevelIdx(1))
                        pos.Y += DrawString(m_font_medium, "Press TAB to move to the next level", pos, Color.LightGray);
                    if (CanMoveLevelIdx(-1))
                        pos.Y += DrawString(m_font_medium, "Press SHIFT-TAB to move to the previous level", pos, Color.LightGray);
                }

                // Show the left button behaviour
                string desc = null;
                switch (m_cam_mode)
                {
                    case ControlMode.MoveCamera:
                        desc = "Rotate the camera";
                        break;

                    case ControlMode.MoveTank:
                        desc = "Move a yellow cannon";
                        break;

                    case ControlMode.AimTank:
                        desc = "Aim any cannon";
                        break;

                    case ControlMode.PickStartTank:
                        desc = "Start by firing a cannon";
                        break;
                }
                if (desc != null)
                {
                    pos.Y += DrawString(m_font_medium, "Left mouse: " + desc, pos, Color.LightBlue);
                }

                // Show the right button behaviour
                if (m_active_level.SimulationStarted)
                {
                    if (m_active_level.SimulationFinished)
                    {
                        pos.Y += DrawString(m_font_medium, "Right mouse: Return to edit mode", pos, Color.LightBlue);
                    }
                    else
                    {
                        pos.Y += DrawString(m_font_medium, "Right mouse: Stop simulation", pos, Color.LightBlue);
                    }
                }

                //// RIGHT SIDE ////

                right_pos.X = graphics.GraphicsDevice.PresentationParameters.BackBufferWidth - 200;

                // Show the current time
                right_pos.Y += DrawTime(m_font, "Time: ", m_active_level.SimulationStarted ? m_active_level.SimulationTime : -1, right_pos, Color.YellowGreen);

                // Show the best time
                right_pos.Y += DrawTime(m_font_medium, "Best: ", m_highscores.GetLevelScore(m_active_level.unique_id), right_pos, Color.YellowGreen);

                //// CENTER NOTES ////

                float center_y = graphics.GraphicsDevice.PresentationParameters.BackBufferHeight * 0.4f;
                if (m_center_note1 != null)
                    center_y += DrawStringCenter(m_font, m_center_note1, center_y, Color.White);

                if (m_center_note2 != null)
                    center_y += DrawStringCenter(m_font_medium, m_center_note2, center_y, Color.LightGray);

                ////

                // Render the version number
                {
                    float ver_y = graphics.GraphicsDevice.PresentationParameters.BackBufferHeight - m_font_medium.LineSpacing - 2;
                    DrawString(m_font_medium, Version, new Vector2(2, ver_y), Color.LightGray);
                }

                m_font_batch.End();
            }

            base.Draw(gameTime);
        }

        private float DrawTime(SpriteFont font, string txt, float time, Vector2 pos, Color color)
        {
            if (time < 0)
                return DrawString(font, txt + "-", pos, color);

            int hsecs = (int)(time * 100.0f);
            int secs = hsecs / 100;
            hsecs %= 100;
            int mins = secs / 60;
            secs %= 60;

            return DrawString(font, String.Format("{0}{1:00}:{2:00}.{3:00}", txt, mins, secs, hsecs), pos, color);
        }

        private float DrawString(SpriteFont font, string txt, Vector2 pos, Color color)
        {
            m_font_batch.DrawString(font, txt, pos + new Vector2(3, 3), Color.Black);
            m_font_batch.DrawString(font, txt, pos, color);
            return font.LineSpacing;
        }

        private float DrawStringCenter(SpriteFont font, string txt, float y, Color color)
        {
            Vector2 size = font.MeasureString(txt);
            float x = (graphics.GraphicsDevice.PresentationParameters.BackBufferWidth - size.X) * 0.5f;

            return DrawString(font, txt, new Vector2(x, y), color);
        }

        private VertexPositionTexture[] m_render_quad_vtx = new VertexPositionTexture[4];

        private void RenderQuad(GraphicsDeviceManager graphics)
        {
            RenderQuad(graphics, 0);
        }

        private void RenderQuad(GraphicsDeviceManager graphics, float z)
        {
            graphics.GraphicsDevice.VertexDeclaration = m_vdecl_VertexPositionTexture;

            m_render_quad_vtx[0] = new VertexPositionTexture(new Vector3(-1, -1, z), new Vector2(0, 1));
            m_render_quad_vtx[1] = new VertexPositionTexture(new Vector3(-1,  1, z), new Vector2(0, 0));
            m_render_quad_vtx[2] = new VertexPositionTexture(new Vector3( 1,  1, z), new Vector2(1, 0));
            m_render_quad_vtx[3] = new VertexPositionTexture(new Vector3( 1, -1, z), new Vector2(1, 1));

            m_normal_shader.CommitChanges();
            m_normal_shader.Begin();
            for (int i = 0; i < m_normal_shader.CurrentTechnique.Passes.Count; ++i)
            {
                EffectPass pass = m_normal_shader.CurrentTechnique.Passes[i];
                pass.Begin();

                graphics.GraphicsDevice.DrawUserPrimitives<VertexPositionTexture>(
                    PrimitiveType.TriangleFan,
                    m_render_quad_vtx,
                    0,
                    2);

                pass.End();
            }
            m_normal_shader.End();
        }

        private void RenderScene(GraphicsDeviceManager graphics)
        {
            m_normal_shader.Begin();
            for (int i = 0; i < m_normal_shader.CurrentTechnique.Passes.Count; ++i)
            {
                EffectPass pass = m_normal_shader.CurrentTechnique.Passes[i];
                pass.Begin();

                RenderScenePass(graphics);

                pass.End();
            }
            m_normal_shader.End();
        }

        private void RenderScenePass(GraphicsDeviceManager graphics)
        {
            Matrix world_mtx = Matrix.Identity;

            world_mtx = Matrix.Identity;
            m_normal_shader.Parameters["World"].SetValue(world_mtx);
            m_normal_shader.Parameters["WorldViewProj"].SetValue(world_mtx * m_view_mtx * m_proj_mtx);
            m_normal_shader.Parameters["LightColor"].SetValue(new Vector4(1, 1, 1, 1));
            m_normal_shader.CommitChanges();

            graphics.GraphicsDevice.VertexDeclaration = m_vdecl_VertexPositionNormalColor;

            // Render the background
            Render(graphics, m_bg_mesh.vtx, m_bg_mesh.idx);

            // Render the field
            m_normal_shader.Parameters["LightColor"].SetValue(new Vector4(0.98f, 0.9f, 0.5f, 1));
            m_normal_shader.CommitChanges();
            Render(graphics, m_active_level);

            // Render the tanks
            for (int idx = 0; idx < m_active_level.tanks.Count; ++idx)
            {
                Tank tank = m_active_level.tanks[idx];

                Vector4 base_color = new Vector4(1, 1, 1, 1);
                Vector4 gun_color = new Vector4(1, 1, 1, 1);

                if (tank.fixed_pos)
                    base_color = new Vector4(10.0f, 1.0f, 1.0f, 1);

                switch (m_cam_mode)
                {
                    case ControlMode.MoveTank:
                        if (tank.fixed_pos)
                        {
                            gun_color = base_color;
                        }
                        else
                        {
                            if (m_moving_tank_idx == -1 || m_moving_tank_idx == idx)
                            {
                                base_color = gun_color = new Vector4(5.0f, 5.0f, 1.0f, 1.0f);
                                if (m_moving_tank_idx != -1 && IsTankAtValidPlace(m_moving_tank_idx) == false)
                                { // Cannot move to this place
                                    base_color = gun_color = new Vector4(15.0f, 0.0f, 0.0f, 1.0f);
                                }
                            }
                        }
                        break;

                    case ControlMode.AimTank:
                        if (m_moving_tank_idx == -1 || m_moving_tank_idx == idx)
                        {
                            base_color = gun_color = new Vector4(1.0f, 1.0f, 5.0f, 1.0f);
                        }
                        break;

                    case ControlMode.PickStartTank:
                        base_color = gun_color = new Vector4(1.0f, 5.0f, 1.0f, 1.0f);
                        break;

                    default:
                        break;
                }

                // Draw the base
                world_mtx = tank.BaseMatrix;
                m_normal_shader.Parameters["World"].SetValue(world_mtx);
                m_normal_shader.Parameters["WorldViewProj"].SetValue(world_mtx * m_view_mtx * m_proj_mtx);
                m_normal_shader.Parameters["LightColor"].SetValue(base_color);
                m_normal_shader.CommitChanges();

                Render(graphics, m_tank_meshes.meshvtx_base, m_tank_meshes.meshidx_base);

                // Draw the guns/top
                Matrix top_mtx = tank.TopMatrix;
                world_mtx = top_mtx;
                m_normal_shader.Parameters["World"].SetValue(world_mtx);
                m_normal_shader.Parameters["WorldViewProj"].SetValue(world_mtx * m_view_mtx * m_proj_mtx);
                m_normal_shader.Parameters["LightColor"].SetValue(base_color);
                m_normal_shader.CommitChanges();

                Render(graphics, m_tank_meshes.meshvtx_top[(int)tank.type], m_tank_meshes.meshidx_top[(int)tank.type]);

                // Draw the remaining amount of bullets on the tank
                int bullets_left = tank.bullets - tank.fired;
                if (bullets_left > 0)
                {
                    float ang_add = 55.0f;
                    float ang = -(ang_add * (bullets_left - 1) * 0.5f);

                    m_normal_shader.Parameters["LightColor"].SetValue(new Vector4(14.0f, 14.0f, 14.0f, 1.0f));

                    for (int i = 0; i < bullets_left; ++i, ang += ang_add)
                    {
                        world_mtx = Matrix.CreateScale(0.7f) * Matrix.CreateTranslation(0, 0.15f, 0.15f) * Matrix.CreateRotationY(MathHelper.ToRadians(ang)) * top_mtx;
                        m_normal_shader.Parameters["World"].SetValue(world_mtx);
                        m_normal_shader.Parameters["WorldViewProj"].SetValue(world_mtx * m_view_mtx * m_proj_mtx);
                        m_normal_shader.CommitChanges();

                        Render(graphics, m_tank_meshes.meshvtx_bullet, m_tank_meshes.meshidx_bullet);
                    }
                }
            }

            // Render the bullets
            for (int i = 0; i < m_active_level.bullets.Count; ++i)
            {
                Vector3 bullet_pos = m_active_level.bullets[i].pos;

                world_mtx = Matrix.CreateTranslation(bullet_pos);
                m_normal_shader.Parameters["World"].SetValue(world_mtx);
                m_normal_shader.Parameters["WorldViewProj"].SetValue(world_mtx * m_view_mtx * m_proj_mtx);
                m_normal_shader.Parameters["LightColor"].SetValue(new Vector4(3, 3, 3, 1));
                m_normal_shader.CommitChanges();

                Render(graphics, m_tank_meshes.meshvtx_bullet, m_tank_meshes.meshidx_bullet);
            }
        }

        private void RenderSceneEffects(GraphicsDeviceManager graphics)
        {
            m_normal_shader.CurrentTechnique = m_normal_shader.Techniques["AlphaRenderTechnique"];

            m_normal_shader.Begin();
            for (int i = 0; i < m_normal_shader.CurrentTechnique.Passes.Count; ++i)
            {
                EffectPass pass = m_normal_shader.CurrentTechnique.Passes[i];
                pass.Begin();

                RenderSceneEffectsPass(graphics);

                pass.End();
            }
            m_normal_shader.End();
        }

        private VertexPositionNormalColor[] m_render_water_vtx = new VertexPositionNormalColor[4];

        private void RenderSceneEffectsPass(GraphicsDeviceManager graphics)
        {
            Matrix world_mtx;

            // Render the tank lines
            graphics.GraphicsDevice.RenderState.BlendFunction = BlendFunction.Add;
            graphics.GraphicsDevice.RenderState.SourceBlend = Blend.One;
            graphics.GraphicsDevice.RenderState.DestinationBlend = Blend.One;
            graphics.GraphicsDevice.RenderState.DepthBufferWriteEnable = false;

            world_mtx = Matrix.Identity;
            m_normal_shader.Parameters["World"].SetValue(world_mtx);
            m_normal_shader.Parameters["WorldViewProj"].SetValue(world_mtx * m_view_mtx * m_proj_mtx);
            m_normal_shader.CommitChanges();

            float aim_line_len = 5.0f;
            if (m_active_level.SimulationStarted)
                aim_line_len = 0.0f;

            for (int idx = 0; idx < m_active_level.tanks.Count; ++idx)
            {
                Tank tank = m_active_level.tanks[idx];

                // DEBUG: Render the gun lines
                for (int gidx = 0; gidx < tank.GunCount; ++gidx)
                {
                    Vector3 gdir = tank.GetGunDir(gidx);
                    RenderLine(graphics, tank.GunPos3D, tank.GunPos3D + gdir * aim_line_len, new Color(255, 0, 0, 255));
                }
                //RenderLine(graphics, tank.GunPos3D, tank.GunPos3D + tank.look_dir * 5.0f, Color.Red);
            }

            // Generate the water vertices
            {
                Color water_col = new Color(10, 10, 234, 100);
                float wy = -0.2f;
                float xz = 4.9f;

                m_render_water_vtx[0].Position = new Vector3(-xz, wy, -xz);
                m_render_water_vtx[2].Position = new Vector3(-xz, wy, xz);
                m_render_water_vtx[1].Position = new Vector3(xz, wy, -xz);
                m_render_water_vtx[3].Position = new Vector3(xz, wy, xz);
                m_render_water_vtx[0].Color = water_col;
                m_render_water_vtx[1].Color = water_col;
                m_render_water_vtx[2].Color = water_col;
                m_render_water_vtx[3].Color = water_col;

                graphics.GraphicsDevice.RenderState.SourceBlend = Blend.SourceAlpha;
                graphics.GraphicsDevice.RenderState.DestinationBlend = Blend.InverseSourceAlpha;

                graphics.GraphicsDevice.DrawUserPrimitives<VertexPositionNormalColor>(
                    PrimitiveType.TriangleStrip,
                    m_render_water_vtx, 0, 2);
            }

            // Render the particles
            if (m_particle_mgr.vertex_count > 0)
            {
                graphics.GraphicsDevice.RenderState.SourceBlend = Blend.SourceAlpha;
                graphics.GraphicsDevice.RenderState.DestinationBlend = Blend.InverseSourceAlpha;
                graphics.GraphicsDevice.RenderState.PointSize = 4.0f;
                graphics.GraphicsDevice.VertexDeclaration = m_vdecl_VertexPositionColor;

                graphics.GraphicsDevice.DrawUserPrimitives<VertexPositionColor>(
                    PrimitiveType.PointList,
                    m_particle_mgr.vertices, 0, m_particle_mgr.vertex_count);

            }

            // Done
            graphics.GraphicsDevice.RenderState.DepthBufferWriteEnable = true;
            graphics.GraphicsDevice.VertexDeclaration = m_vdecl_VertexPositionNormalColor;
        }
            

        private void Render(GraphicsDeviceManager graphics, ActiveLevel level)
        {
            if (level.Triangles.Count == 0)
                return;

            graphics.GraphicsDevice.DrawUserIndexedPrimitives<VertexPositionNormalColor>(
                PrimitiveType.TriangleList,
                level.Vertices.RawList, 0, level.Vertices.Count,
                level.Triangles.RawList, 0, level.Triangles.Count / 3);
        }

        private void Render(GraphicsDeviceManager graphics, VertexPositionNormalColor[] vtx, int[] idx)
        {
            if (idx.Length == 0)
                return;

            graphics.GraphicsDevice.DrawUserIndexedPrimitives<VertexPositionNormalColor>(
                PrimitiveType.TriangleList,
                vtx, 0, vtx.Length,
                idx, 0, idx.Length / 3);
        }

        private VertexPositionNormalColor[] m_render_line_vtx = new VertexPositionNormalColor[2];

        private void RenderLine(GraphicsDeviceManager graphics, Vector3 start, Vector3 end, Color color)
        {
            m_render_line_vtx[0].Position = start;
            m_render_line_vtx[0].Color = color;
            m_render_line_vtx[1].Position = end;
            m_render_line_vtx[1].Color = color;

            graphics.GraphicsDevice.DrawUserPrimitives<VertexPositionNormalColor>(
                PrimitiveType.LineList,
                m_render_line_vtx,
                0, 1);
        }
    }
}
